package flow.components
import chisel3._
import chisel3.util.experimental.decode.{decoder, TruthTable}
import chisel3.util.{Valid, log2Ceil}
import chisel3.util.MuxLookup
import chisel3.util.BitPat
import shapeless.{HNil, ::}
import shapeless.HList
import flow.Params
import RV32InstSubfields._
import flow.components.util._

class PcControlInterface extends Bundle {
  object SrcSelect extends ChiselEnum {
    val pStaticNpc, pExeOut = Value
  }

  val useImmB = Input(Bool())
  val srcSelect = Input(SrcSelect())

  def ctrlBindPorts = {
    useImmB :: srcSelect :: HNil
  }
}

class ProgramCounter[T <: UInt](tpe: T) extends Module {

  val control = IO(new PcControlInterface)
  val in = IO(new Bundle {
    val immB = Input(tpe)
    val exeOut = Input(tpe)
  })
  val out = IO(Output(tpe))

  private val pc_reg = RegInit(0x80000000L.U)

//   pc := in.pcSrcs(control.srcSelect.asUInt)
  import control.SrcSelect._
  when(control.useImmB === true.B) {
    pc_reg := pc_reg + in.immB
  }.elsewhen(control.srcSelect === pStaticNpc) {
    pc_reg := pc_reg + 4.U
  }.elsewhen(control.srcSelect === pExeOut) {
    pc_reg := in.exeOut
  }
  out := pc_reg
}

object ProgramCounter {
  def apply[T <: UInt](tpe: T): ProgramCounter[T] = {
    val pc = Module(new ProgramCounter(tpe))
    pc
  }
}

object newPcControlInterface {
  object SrcSelect extends ChiselEnum {
    val pStatic, pJmp, pBR = Value
  }

  class newPcControlInterface extends Bundle {
    val srcSelect = Input(SrcSelect())

    def ctrlBindPorts = {
      srcSelect :: HNil
    }
  }
  def apply() = new newPcControlInterface;
}

class newProgramCounter(implicit p: Params) extends Module {
  val control = IO(newPcControlInterface())
  import newPcControlInterface.SrcSelect._
  val in = IO(new Bundle {
    val brOffset = Input(UInt(p.XLEN))
    val jAddr = Input(UInt(p.XLEN))
  })
  val out = IO(Output(UInt(p.XLEN)))

  private val pcReg = RegInit(p.resetVector.U)
  out := pcReg

  private val npc = MuxLookup(control.srcSelect, 4.U)(
    Seq(
      pStatic -> (pcReg + 4.U),
      pJmp -> in.jAddr,
      pBR -> (pcReg + in.brOffset)
    )
  )

  pcReg := npc
}

class PcController(implicit p: Params) extends Module {
  val in = IO(new Bundle {
    val inst = Input(UInt(p.instWidth))
  })

  val out = IO(Flipped(newPcControlInterface()))

  import RV32Inst._
  import newPcControlInterface.SrcSelect._
  private val _jmpMapping = Array(jal, jalr).map(_ -> pJmp.BP)
  private val _brMapping =
    Array(beq, bne, blt, bge, bltu, bgeu).map(_ -> pBR.BP)

  val mapping = TruthTable(
    _jmpMapping ++ _brMapping,
    pStatic.BP
  )

  out.srcSelect := newPcControlInterface.SrcSelect
    .safe(
      (
        decoder(in.inst, mapping)
      )
    )
    ._1
}
